"""
Tests for the SteadyText Command-Line Interface (CLI).

AIDEV-NOTE: These tests use click.testing.CliRunner to invoke CLI commands
and verify their output and behavior.
"""

import json
import pytest
from click.testing import CliRunner
from steadytext.cli.main import cli
from steadytext import __version__


@pytest.fixture
def runner():
    """Fixture for invoking command-line calls."""
    return CliRunner()


class TestCli:
    """Test suite for the main CLI and generate command."""

    def test_cli_version(self, runner):
        """Test that `st --version` returns the correct version."""
        result = runner.invoke(cli, ["--version"])
        assert result.exit_code == 0
        assert __version__ in result.output

    def test_generate_stdin(self, runner):
        """Test `st` command with stdin."""
        result = runner.invoke(cli, input="Hello")
        assert result.exit_code == 0
        assert len(result.output) > 0

    def test_generate_command_basic(self, runner):
        """Test `st generate` command."""
        result = runner.invoke(cli, ["generate", "A test prompt"])
        assert result.exit_code == 0
        assert len(result.output) > 0

    def test_generate_wait_mode(self, runner):
        """Test `st generate --wait`."""
        result = runner.invoke(cli, ["generate", "A test prompt", "--wait"])
        assert result.exit_code == 0
        assert len(result.output) > 0
        # In fallback mode, streaming vs. wait might not differ, but we test the flag is accepted.

    def test_generate_json_output(self, runner):
        """Test `st generate --json`."""
        result = runner.invoke(cli, ["generate", "A test prompt", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert "text" in data
        assert "model" in data
        assert "usage" in data
        assert isinstance(data["text"], str)

    def test_generate_with_size(self, runner):
        """Test `st generate --size`."""
        result = runner.invoke(cli, ["generate", "A test prompt", "--size", "small"])
        assert result.exit_code == 0
        assert len(result.output) > 0

    def test_generate_with_invalid_size(self, runner):
        """Test `st generate` with an invalid size."""
        result = runner.invoke(cli, ["generate", "A test prompt", "--size", "tiny"])
        assert result.exit_code != 0
        assert "Invalid value for '--size'" in result.output

    def test_generate_with_logprobs(self, runner):
        """Test `st generate --logprobs`."""
        result = runner.invoke(
            cli, ["generate", "A test prompt", "--logprobs", "--json"]
        )
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert "logprobs" in data
        # In fallback mode, logprobs will be None
        assert data["logprobs"] is None


class TestEmbedCli:
    """Test suite for the `embed` CLI command."""

    def test_embed_basic(self, runner):
        """Test `st embed` with a single string."""
        result = runner.invoke(cli, ["embed", "some text to embed"])
        assert result.exit_code == 0
        # Default output is a hex-encoded numpy array
        assert len(result.output.strip()) > 0

    def test_embed_json_output(self, runner):
        """Test `st embed --json`."""
        result = runner.invoke(cli, ["embed", "some text", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert "embedding" in data
        assert "model" in data
        assert "usage" in data
        assert isinstance(data["embedding"], list)
        assert len(data["embedding"]) == 1024

    def test_embed_numpy_output(self, runner):
        """Test `st embed --numpy`."""
        result = runner.invoke(cli, ["embed", "some text", "--numpy"])
        assert result.exit_code == 0
        # Output should be a string representation of a numpy array
        assert result.output.strip().startswith("[")
        assert result.output.strip().endswith("]")

    def test_embed_multiple_strings(self, runner):
        """Test `st embed` with multiple strings."""
        result = runner.invoke(cli, ["embed", "text one", "text two", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert len(data["embedding"]) == 1024


class TestCacheCli:
    """Test suite for the `cache` CLI command."""

    @pytest.fixture(autouse=True)
    def enable_cache_for_tests(self, monkeypatch):
        """Temporarily enable cache initialization for these tests."""
        monkeypatch.delenv("STEADYTEXT_SKIP_CACHE_INIT", raising=False)

    def test_cache_path(self, runner):
        """Test `st cache path`."""
        result = runner.invoke(cli, ["cache", "path"])
        assert result.exit_code == 0
        assert "steadytext" in result.output
        assert "caches" in result.output

    def test_cache_status(self, runner):
        """Test `st cache stats`."""
        result = runner.invoke(cli, ["cache", "stats"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert "generation" in data
        assert "embedding" in data

    def test_cache_clear(self, runner):
        """Test `st cache clear`."""
        # First, add something to the cache
        runner.invoke(cli, ["generate", "populate cache for clear test"])

        # Now, clear it
        result = runner.invoke(cli, ["cache", "clear", "--yes"])
        assert result.exit_code == 0
        assert "All caches cleared" in result.output

        # Verify it's empty
        status_result = runner.invoke(cli, ["cache", "stats"])
        data = json.loads(status_result.output)
        assert data["generation"]["size"] == 0
        assert data["embedding"]["size"] == 0


class TestModelsCli:
    """Test suite for the `models` CLI command."""

    def test_models_list(self, runner):
        """Test `st models list`."""
        result = runner.invoke(cli, ["models", "list"])
        assert result.exit_code == 0
        assert "gemma-3n-2b" in result.output
        assert "gemma-3n-4b" in result.output

    def test_models_path(self, runner):
        """Test `st models path`."""
        result = runner.invoke(cli, ["models", "path"])
        assert result.exit_code == 0
        assert "steadytext" in result.output
        assert "models" in result.output

    def test_models_preload(self, runner, monkeypatch):
        """Test `st models preload`."""
        # This just checks that the command runs without error, as preloading
        # is disabled in tests.
        result = runner.invoke(
            cli, ["models", "preload"], env={"STEADYTEXT_SKIP_MODEL_LOAD": "1"}
        )
        assert result.exit_code == 0
        assert "Preloading models... (skipped in test environment)" in result.output
