import json
import numpy as np
from click.testing import CliRunner

from steadytext import generate, embed
from steadytext.cli.main import cli


def test_generate_default_seed():
    output1 = generate("test")
    output2 = generate("test", seed=42)
    assert output1 == output2


def test_embed_default_seed():
    output1 = embed("test")
    output2 = embed("test", seed=42)
    assert np.array_equal(output1, output2)


def test_generate_custom_seed_determinism():
    output1 = generate("test", seed=123)
    output2 = generate("test", seed=123)
    assert output1 == output2


def test_embed_custom_seed_determinism():
    output1 = embed("test", seed=123)
    output2 = embed("test", seed=123)
    assert np.array_equal(output1, output2)


def test_generate_different_seeds():
    output1 = generate("test", seed=123)
    output2 = generate("test", seed=456)
    assert output1 != output2


def test_embed_different_seeds():
    output1 = embed("test", seed=123)
    output2 = embed("test", seed=456)
    assert not np.array_equal(output1, output2)


def test_cli_generate_default_seed():
    runner = CliRunner()
    result1 = runner.invoke(cli, ["generate", "test"])
    result2 = runner.invoke(cli, ["generate", "test", "--seed", "42"])
    assert result1.stdout == result2.stdout


def test_cli_generate_custom_seed():
    runner = CliRunner()
    result1 = runner.invoke(cli, ["generate", "test", "--seed", "123"])
    result2 = runner.invoke(cli, ["generate", "test", "--seed", "123"])
    assert result1.stdout == result2.stdout


def test_cli_generate_different_seeds():
    runner = CliRunner()
    result1 = runner.invoke(cli, ["generate", "test", "--seed", "123"])
    result2 = runner.invoke(cli, ["generate", "test", "--seed", "456"])
    assert result1.stdout != result2.stdout


def test_cli_embed_default_seed():
    runner = CliRunner()
    result1 = runner.invoke(cli, ["embed", "test", "--json"])
    result2 = runner.invoke(cli, ["embed", "test", "--json", "--seed", "42"])
    # Parse JSON and compare without time_taken
    json1 = json.loads(result1.stdout)
    json2 = json.loads(result2.stdout)
    json1.pop("time_taken", None)
    json2.pop("time_taken", None)
    assert json1 == json2


def test_cli_embed_custom_seed():
    runner = CliRunner()
    result1 = runner.invoke(cli, ["embed", "test", "--json", "--seed", "123"])
    result2 = runner.invoke(cli, ["embed", "test", "--json", "--seed", "123"])
    # Parse JSON and compare without time_taken
    json1 = json.loads(result1.stdout)
    json2 = json.loads(result2.stdout)
    json1.pop("time_taken", None)
    json2.pop("time_taken", None)
    assert json1 == json2


def test_cli_embed_different_seeds():
    runner = CliRunner()
    result1 = runner.invoke(cli, ["embed", "test", "--json", "--seed", "123"])
    result2 = runner.invoke(cli, ["embed", "test", "--json", "--seed", "456"])
    # Parse JSON and compare embeddings
    json1 = json.loads(result1.stdout)
    json2 = json.loads(result2.stdout)
    assert json1["embedding"] != json2["embedding"]


# Edge case tests
def test_negative_seed_validation():
    """Test that negative seeds raise ValueError"""
    import pytest
    from steadytext import generate, embed
    
    with pytest.raises(ValueError, match="Seed must be a non-negative integer"):
        generate("test", seed=-1)
    
    with pytest.raises(ValueError, match="Seed must be a non-negative integer"):
        embed("test", seed=-5)


def test_large_seed_values():
    """Test that large seed values work correctly"""
    from steadytext import generate, embed
    
    # Test with large seed values
    large_seed = 2**31 - 1  # Maximum 32-bit integer
    text1 = generate("test", seed=large_seed)
    embedding1 = embed("test", seed=large_seed)
    
    # Should not raise errors and should be deterministic
    text2 = generate("test", seed=large_seed)
    embedding2 = embed("test", seed=large_seed)
    
    assert text1 == text2
    assert np.array_equal(embedding1, embedding2)


def test_seed_type_validation():
    """Test that non-integer seeds raise appropriate errors"""
    import pytest
    from steadytext import generate, embed
    
    # Test with float
    with pytest.raises(ValueError, match="Seed must be a non-negative integer"):
        generate("test", seed=3.14)
    
    # Test with string
    with pytest.raises(ValueError, match="Seed must be a non-negative integer"):
        embed("test", seed="42")
    
    # Test with None
    with pytest.raises(ValueError, match="Seed must be a non-negative integer"):
        generate("test", seed=None)
